//! Storage layer trait
use crate::{
    amount::AmountCents, payment::PaymentTo, transaction::Type, AccountId, Amount, Asset, BaseTx,
    PaymentFrom, PaymentId, RevId, Revision, Transaction, TxId,
};
//use chrono::{DateTime, Utc};
use serde::Serialize;

pub mod cache;
#[cfg(any(feature = "sqlite", test))]
pub mod sqlite;
pub use self::cache::Cache;
#[cfg(any(feature = "sqlite", test))]
pub use self::sqlite::SQLite;

#[derive(Clone, Copy, PartialEq, Eq, Debug, Serialize)]
/// Storage Payment Status
pub enum ReceivedPaymentStatus {
    ///  The payment is available to be spent
    Spendable,
    /// The payment is not spentable and will not be part of the balance
    Failed,
    /// The payment is unspendable temporarily, it can be updated to any other status
    Locked,
    /// The transaction has been spent
    Spent,
}

/// Serializes an object to bytes using the default serialization method
pub fn to_bytes<T>(val: &T) -> Result<Vec<u8>, Error>
where
    T: serde::Serialize + borsh::BorshSerialize,
{
    borsh::to_vec(val).map_err(|e| Error::Encoding(e.to_string()))
}

/// Deserializes an object from bytes using the default deserialization method
pub fn from_bytes<T>(val: &[u8]) -> Result<T, Error>
where
    T: serde::de::DeserializeOwned + borsh::BorshDeserialize,
{
    borsh::BorshDeserialize::deserialize(&mut val.as_ref())
        .map_err(|e| Error::Encoding(e.to_string()))
}

impl From<ReceivedPaymentStatus> for u32 {
    fn from(val: ReceivedPaymentStatus) -> Self {
        match val {
            ReceivedPaymentStatus::Spendable => 0,
            ReceivedPaymentStatus::Locked => 20,
            ReceivedPaymentStatus::Spent => 30,
            ReceivedPaymentStatus::Failed => 40,
        }
    }
}

impl TryFrom<u32> for ReceivedPaymentStatus {
    type Error = Error;

    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(ReceivedPaymentStatus::Spendable),
            20 => Ok(ReceivedPaymentStatus::Locked),
            30 => Ok(ReceivedPaymentStatus::Spent),
            40 => Ok(ReceivedPaymentStatus::Failed),
            _ => Err(Error::Storage(format!("Invalid status: {}", value))),
        }
    }
}

/// IDs to query transactions
#[derive(Debug, Clone)]
pub enum FilterId {
    /// By transaction ID with the latest revision
    LatestRevision(TxId),
    /// A transaction at a given revision
    RevisionId(RevId),
}

impl From<TxId> for FilterId {
    fn from(id: TxId) -> Self {
        FilterId::LatestRevision(id)
    }
}

impl From<&TxId> for FilterId {
    fn from(id: &TxId) -> Self {
        FilterId::LatestRevision(id.clone())
    }
}

impl From<RevId> for FilterId {
    fn from(id: RevId) -> Self {
        FilterId::RevisionId(id)
    }
}

impl From<&RevId> for FilterId {
    fn from(id: &RevId) -> Self {
        FilterId::RevisionId(id.clone())
    }
}

impl ToString for FilterId {
    fn to_string(&self) -> String {
        match self {
            FilterId::LatestRevision(id) => id.to_string(),
            FilterId::RevisionId(id) => id.to_string(),
        }
    }
}

#[derive(thiserror::Error, Debug, Serialize)]
/// Storage error
pub enum Error {
    #[error("Storage error: {0}")]
    /// Any error related to the storage layer, with string
    Storage(String),

    #[error("I/O error: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    /// Any error related to the storage layer, with std::io::Error
    Io(#[from] std::io::Error),

    #[error("Invalid date format: {0}")]
    /// The date format is invalid
    InvalidDate(String),

    #[error("Invalid id: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    /// Invalid Id
    InvalidId(#[from] crate::id::Error),

    #[error("Spend payment: {0}")]
    /// The requested payment has been spent already
    SpendPayment(String),

    #[error("No storage update when expecting")]
    /// No update was performed when expecting one
    NoUpdate,

    #[error("Math error, overflow")]
    /// Overflow error
    Overflow,

    #[error("Math error, underflow")]
    /// Underflow error
    Underflow,

    #[error("Record not found")]
    /// The requested record was not found
    NotFound,

    #[error("Internal error with encoding: {0}")]
    /// Error with the encoding/decoding
    Encoding(String),

    #[error("Not enough unspent payments (missing {0} cents)")]
    /// TODO: Convert the AmountCents to Amount for better error reporting upstream
    NotEnoughUnspentPayments(AmountCents),
}

#[async_trait::async_trait]
/// Batches
///
/// Batches are a group of updates to the databases, in which all of the are
/// executed or none. Same concept as a rdbms transaction.
pub trait Batch<'a> {
    /// The current batch is discarded
    async fn rollback(self) -> Result<(), Error>;

    /// The batch is commited into the storage layer and it is consumed.
    async fn commit(self) -> Result<(), Error>;

    /// Flag the given payments as spent by the given transaction.
    async fn spend_payments(
        &mut self,
        transaction_id: &TxId,
        spends: Vec<PaymentId>,
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error>;

    /// Create a new list of payments
    async fn create_payments(
        &mut self,
        transaction_id: &TxId,
        recipients: &[PaymentTo],
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error>;

    /// Updates all payments related to a transaction
    ///
    /// This function will update the created and spent payments related to a given transaction.
    ///
    /// This function must check the previous status before changing it. For instance, Failed and
    /// Spent are final statuses, making the transaction read-only, going forward. The status
    /// from NotSpentable to Spendable can only be done by the parent transaction, and not other.
    async fn update_transaction_payments(
        &mut self,
        transaction_id: &TxId,
        create_status: ReceivedPaymentStatus,
        spend_status: ReceivedPaymentStatus,
    ) -> Result<(usize, usize), Error>;

    /// Returns the stats of a payment from the point of view of the on-going transaction
    async fn get_payment_status(
        &mut self,
        transaction_id: &TxId,
    ) -> Result<Option<ReceivedPaymentStatus>, Error>;

    /// Stores a transaction
    async fn store_base_transaction(
        &mut self,
        transaction_id: &TxId,
        transaction: &BaseTx,
    ) -> Result<(), Error>;

    /// Persists a revision
    async fn store_revision(
        &mut self,
        revision_id: &RevId,
        revision: &Revision,
    ) -> Result<(), Error>;

    /// Updates the revision of a transaction. Both the revision and the transaction must exist.
    async fn update_transaction_revision(
        &mut self,
        transaction_id: &TxId,
        revision_id: &RevId,
        previous_rev_id: Option<&RevId>,
    ) -> Result<(), Error>;

    /// Sets the tags for a given transaction. Any tag not included in this
    /// vector should be removed
    async fn tag_transaction(
        &mut self,
        transaction_id: &TxId,
        tags: &[String],
    ) -> Result<(), Error>;

    /// Creates a relationship between an account and a transaction.
    ///
    /// This is a chance to build an index which relates accounts to
    /// transactions (spending or receiving accounts). This information is used
    /// to query ledger information per account and to enhance any future cache
    /// layer.
    async fn relate_account_to_transaction(
        &mut self,
        transaction: &TxId,
        account: &AccountId,
        typ: Type,
    ) -> Result<(), Error>;
}

#[async_trait::async_trait]
/// Main storage layer
pub trait Storage {
    /// The batch trait.
    ///
    /// This batch is needed to perform the changes in the transactions and
    /// payments in a atomic way
    type Batch<'a>: Batch<'a> + Send
    where
        Self: 'a;

    /// Begins a transaction
    ///
    /// A transaction prepares the storage for a batch of operations, where all
    /// the operations are persisted or none.
    ///
    /// The transaction is returned as a Batch object, which is used to perform
    /// the changes in the transactions and payments.
    ///
    /// The batch has also a rollback
    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error>;

    /// Returns the balances for a given account
    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error>;

    /// Returns a list of unspent payments for a given account and asset.
    ///
    /// The payments should be returned sorted by ascending amount, this bit is
    /// important to make sure that all negative payments are always taken into
    /// account. It will also improve the database by using many small payments
    /// instead of a few large ones, which will make the database faster leaving
    /// fewer unspent payments when checking for balances.
    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error>;

    /*
    ///
    async fn get_revision(&self, revision_id: &TxId) -> Result<Transaction, Error>;

    ///
    async fn get_all_revisions(
        &self,
        transaction_id: &TxId,
    ) -> Result<Vec<(TxId, String, DateTime<Utc>)>, Error>;
    */

    /// Returns a transaction by id, the transaction is returned as a tuple of base transaction and
    /// the current revision
    async fn get_transaction_and_revision(
        &self,
        transaction_id: FilterId,
    ) -> Result<(BaseTx, Revision), Error>;

    /// List all revisions for a given transaction
    async fn get_revisions(&self, transaction_id: &TxId) -> Result<Vec<RevId>, Error>;

    /// Returns a list of a transactions for a given account (and optionally
    /// filter by types). The list of transactions is expected to be sorted by
    /// date desc (newest transactions first),
    async fn get_transactions_with_latest_revision(
        &self,
        account: &AccountId,
        types: &[Type],
        tags: &[String],
    ) -> Result<Vec<(BaseTx, Revision)>, Error>;

    /// Returns a revision with a transaction object by id
    async fn get_transaction(&self, id: FilterId) -> Result<Transaction, Error> {
        let (base_tx, revision) = self.get_transaction_and_revision(id).await?;
        (
            self.get_revisions(&revision.transaction_id).await?,
            base_tx,
            revision,
        )
            .try_into()
            .map_err(|e: crate::transaction::Error| Error::Encoding(e.to_string()))
    }

    /// Get list of transactions
    async fn get_transactions(
        &self,
        account: &AccountId,
        types: &[Type],
        tags: &[String],
    ) -> Result<Vec<Transaction>, Error> {
        let mut transactions = Vec::new();
        for (base_tx, revision) in self
            .get_transactions_with_latest_revision(account, types, tags)
            .await?
            .into_iter()
        {
            transactions.push(
                (
                    self.get_revisions(&revision.transaction_id).await?,
                    base_tx,
                    revision,
                )
                    .try_into()
                    .map_err(|e: crate::transaction::Error| Error::Encoding(e.to_string()))?,
            );
        }
        Ok(transactions)
    }
}

#[cfg(test)]
pub mod test {
    use super::*;
    use crate::{config::Config, status::StatusManager, Transaction};
    use rand::Rng;

    #[macro_export]
    macro_rules! storage_unit_test {
        ($name:ident) => {
            #[tokio::test]
            async fn $name() {
                let ledger = get_instance(stringify!($name)).await;
                $crate::storage::test::$name(ledger).await;
            }
        };
    }

    #[macro_export]
    macro_rules! storage_test_suite {
        () => {
            $crate::storage_unit_test!(transaction);
            $crate::storage_unit_test!(transaction_does_not_update_stale_transactions);
            $crate::storage_unit_test!(transaction_not_available_until_commit);
            $crate::storage_unit_test!(sorted_unspent_payments);
            $crate::storage_unit_test!(does_not_update_spent_payments);
            $crate::storage_unit_test!(does_not_spend_unspendable_payments);
            $crate::storage_unit_test!(spend_spendable_payments);
            $crate::storage_unit_test!(relate_account_to_transaction);
            $crate::storage_unit_test!(not_spendable_new_payments_not_spendable);
        };
    }

    pub async fn transaction_does_not_update_stale_transactions<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let config = Config {
            storage,
            status: StatusManager::default(),
        };

        let asset: Asset = "USD/2".parse().expect("valid asset");
        let mut pending = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "pending".into(),
            vec![(
                "alice".parse().expect("account"),
                asset.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");
        pending.persist(&config).await.expect("valid insert");

        pending
            .clone()
            .change_status(&config, "processing".into(), "some text".to_owned())
            .await
            .expect("valid update");

        pending
            .change_status(
                &config,
                "failed".into(),
                "update from pending to failed (which is not the latest state and should fail)"
                    .to_owned(),
            )
            .await
            .expect_err("stale updates are rejected by storage");
    }

    pub async fn transaction<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let asset: Asset = "USD/2".parse().expect("valid asset");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![(
                "alice".parse().expect("account"),
                asset.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.rollback().await.expect("rollback");
        assert!(storage.get_transaction((&deposit.id).into()).await.is_err());

        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.commit().await.expect("commit");
        assert!(storage.get_transaction((&deposit.id).into()).await.is_ok());
    }

    pub async fn transaction_not_available_until_commit<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![(
                "alice".parse().expect("account"),
                usd.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        assert!(storage.get_transaction((&deposit.id).into()).await.is_err());
        storing.rollback().await.expect("rollback");
        assert!(storage.get_transaction((&deposit.id).into()).await.is_err());

        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store base tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.commit().await.expect("commit");
        assert!(storage.get_transaction((&deposit.id).into()).await.is_ok());
    }

    pub async fn does_not_update_spent_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        let transaction_id = vec![0u8; 32].try_into().expect("valid tx id");

        let recipient = PaymentTo {
            to: account.clone(),
            amount,
        };

        let recipients = vec![recipient];

        writer
            .create_payments(&transaction_id, &recipients, ReceivedPaymentStatus::Locked)
            .await
            .expect("valid payment");

        // Alter state
        assert_eq!(
            writer
                .update_transaction_payments(
                    &transaction_id,
                    ReceivedPaymentStatus::Locked,
                    ReceivedPaymentStatus::Locked
                )
                .await
                .expect("valid"),
            (1, 0)
        );

        // Transactions is settled and their spent payments are forever spent
        assert_eq!(
            writer
                .update_transaction_payments(
                    &transaction_id,
                    ReceivedPaymentStatus::Spent,
                    ReceivedPaymentStatus::Spent
                )
                .await
                .expect("valid"),
            (1, 0)
        );

        for status in [
            ReceivedPaymentStatus::Failed,
            ReceivedPaymentStatus::Spendable,
        ] {
            assert_eq!(
                writer
                    .update_transaction_payments(&transaction_id, status, status)
                    .await
                    .expect("valid update"),
                (0, 0),
                "status: {:?}",
                status
            );
        }
    }

    pub async fn spend_spendable_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        for (i, status) in [
            ReceivedPaymentStatus::Locked,
            ReceivedPaymentStatus::Spendable,
        ]
        .into_iter()
        .enumerate()
        {
            let transaction_id: TxId = vec![i as u8; 32].try_into().expect("valid tx id");

            writer
                .create_payments(
                    &transaction_id,
                    &[PaymentTo {
                        to: account.clone(),
                        amount: amount.clone(),
                    }],
                    status,
                )
                .await
                .expect("valid payment");

            for (spend_new_status, create_new_status) in
                [(ReceivedPaymentStatus::Spent, ReceivedPaymentStatus::Spent)]
            {
                assert_eq!(
                    writer
                        .update_transaction_payments(
                            &transaction_id,
                            spend_new_status,
                            create_new_status
                        )
                        .await
                        .expect("no updates"),
                    (1, 0)
                );
            }
        }
    }

    pub async fn does_not_spend_unspendable_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        for (i, status) in [ReceivedPaymentStatus::Failed, ReceivedPaymentStatus::Spent]
            .into_iter()
            .enumerate()
        {
            let transaction_id: TxId = vec![i as u8; 32].try_into().expect("valid tx id");

            writer
                .create_payments(
                    &transaction_id,
                    &[PaymentTo {
                        to: account.clone(),
                        amount: amount.clone(),
                    }],
                    status,
                )
                .await
                .expect("valid payment");

            for (spend_new_status, create_new_status) in [
                (ReceivedPaymentStatus::Spent, ReceivedPaymentStatus::Spent),
                (
                    ReceivedPaymentStatus::Spendable,
                    ReceivedPaymentStatus::Spendable,
                ),
            ] {
                assert_eq!(
                    writer
                        .update_transaction_payments(
                            &transaction_id,
                            spend_new_status,
                            create_new_status
                        )
                        .await
                        .expect("no updates"),
                    (0, 0)
                );
            }
        }
    }

    pub async fn sorted_unspent_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let accounts: Vec<AccountId> = (0..10)
            .map(|i| format!("account{}", i).parse().expect("account"))
            .collect();

        let mut rng = rand::thread_rng();
        let target_inputs_per_account = 20;
        let usd: Asset = "USD/2".parse().expect("valid asset");

        for (index, account) in accounts.iter().enumerate() {
            let transaction_id: TxId = vec![index as u8; 32].try_into().expect("valid tx id");
            let recipients = (0..target_inputs_per_account)
                .map(|_| {
                    let amount = usd
                        .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
                        .expect("valid amount");
                    PaymentTo {
                        to: account.clone(),
                        amount,
                    }
                })
                .collect::<Vec<_>>();

            writer
                .create_payments(
                    &transaction_id,
                    &recipients,
                    ReceivedPaymentStatus::Spendable,
                )
                .await
                .expect("valid payment");
        }
        writer.commit().await.expect("valid commit");

        let mut at_least_one_negative_amount = false;
        for account in &accounts {
            let balance = &storage.get_balance(account).await.expect("valid balance")[0];
            if balance.cents() < 0 {
                continue;
            }
            let all_unspent_amounts = storage
                .get_unspent_payments(account, balance.asset(), balance.cents())
                .await
                .expect("valid unspent payments")
                .into_iter()
                .map(|x| x.amount.cents())
                .collect::<Vec<_>>();

            let mut sorted = all_unspent_amounts.clone();
            sorted.sort();

            assert_eq!(target_inputs_per_account, all_unspent_amounts.len());
            assert_eq!(sorted, all_unspent_amounts);
            if !at_least_one_negative_amount {
                at_least_one_negative_amount = sorted[0] < 0;
            }
        }
        assert!(at_least_one_negative_amount);
    }

    pub async fn relate_account_to_transaction<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let account1: AccountId = "alice1".parse().expect("account");
        let account2: AccountId = "alice2".parse().expect("account");
        let account3: AccountId = "alice3".parse().expect("account");
        let usd: Asset = "USD/2".parse().expect("USD");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![(
                account1.clone(),
                usd.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");

        let mut batch = storage.begin().await.expect("valid tx");
        batch
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("is ok");
        batch
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        batch
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        batch
            .relate_account_to_transaction(&deposit.id, &account1, Type::Deposit)
            .await
            .expect("relationship");
        batch
            .relate_account_to_transaction(&deposit.id, &account2, Type::Deposit)
            .await
            .expect("relationship");

        batch.commit().await.expect("valid commit");

        assert_eq!(
            1,
            storage
                .get_transactions(&account1, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            1,
            storage
                .get_transactions(&account2, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            0,
            storage
                .get_transactions(&account3, &[Type::Deposit], &[])
                .await
                .expect("valid tx")
                .len()
        );
        for account in &[account1, account2, account3] {
            assert_eq!(
                0,
                storage
                    .get_transactions(account, &[Type::Withdrawal], &[])
                    .await
                    .expect("valid tx")
                    .len()
            );
        }
    }

    pub async fn not_spendable_new_payments_not_spendable<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(10.0..10000.0)))
            .expect("valid amount");

        let transaction: TxId = vec![0u8; 32].try_into().expect("valid tx id");

        writer
            .create_payments(
                &transaction,
                &[PaymentTo {
                    to: account.clone(),
                    amount,
                }],
                ReceivedPaymentStatus::Locked,
            )
            .await
            .expect("valid payment");

        writer.commit().await.expect("is ok");

        assert!(storage
            .get_balance(&account)
            .await
            .expect("valid balance")
            .is_empty());
    }
}
