use crate::{
    config::Config,
    payment::PaymentTo,
    status::InternalStatus,
    storage::{Batch, ReceivedPaymentStatus, Storage},
    AccountId, Amount, PaymentFrom, RevId, Status, TxId,
};
use chrono::{DateTime, TimeZone, Utc};
use serde::{Deserialize, Serialize};
use std::ops::Deref;

mod base_tx;
mod error;
mod revision;
mod typ;

pub use self::{base_tx::BaseTx, error::Error, revision::Revision, typ::Type};

pub(crate) fn to_ts_microseconds<W: std::io::Write>(
    dt: &DateTime<Utc>,
    writer: &mut W,
) -> std::io::Result<()> {
    borsh::BorshSerialize::serialize(&dt.timestamp_millis(), writer)?;
    Ok(())
}

pub(crate) fn from_ts_microseconds<R: borsh::io::Read>(
    reader: &mut R,
) -> ::core::result::Result<DateTime<Utc>, borsh::io::Error> {
    match Utc.timestamp_millis_opt(borsh::BorshDeserialize::deserialize_reader(reader)?) {
        chrono::LocalResult::Single(dt) => Ok(dt.with_timezone(&Utc)),
        _ => Err(borsh::io::Error::new(
            borsh::io::ErrorKind::InvalidData,
            "invalid timestamp".to_owned(),
        )),
    }
}

#[derive(Debug, Clone, Deserialize, Serialize)]
/// A transaction with a revision
///
/// This is the public representation of a transaction, with the revision. Since transactions are
/// immutable, the revision is the only way to update them.
///
/// A Revision is an ammenment to a transaction, but it can only ammend a few fields of the
/// transaction, mainly their status and certain unrelated fields to the transactions, such as tags.
///
/// Revisions are only updatable until the transaction is finalized, by which point they are also
/// immutable.
///
/// Having a revision also makes concurrency easier, since only the latest revision can may be
/// upgraded, any other attempt to ammend a stale revision will be rejected.
pub struct Transaction {
    /// Transaction ID
    #[serde(rename = "_id")]
    pub id: TxId,

    /// List of all revisions
    #[serde(rename = "_all_revs")]
    pub revisions: Vec<RevId>,

    /// Revision ID
    #[serde(rename = "_rev")]
    pub revision_id: RevId,

    /// The revision struct.
    #[serde(flatten)]
    pub revision: Revision,

    #[serde(flatten)]
    /// The transaction struct
    pub transaction: BaseTx,
}

impl Deref for Transaction {
    type Target = BaseTx;

    fn deref(&self) -> &Self::Target {
        &self.transaction
    }
}

impl TryFrom<(Vec<RevId>, BaseTx, Revision)> for Transaction {
    type Error = Error;

    fn try_from(value: (Vec<RevId>, BaseTx, Revision)) -> Result<Self, Self::Error> {
        Ok(Self {
            id: value.1.id()?,
            revisions: value.0,
            revision_id: value.2.rev_id()?,
            transaction: value.1,
            revision: value.2,
        })
    }
}

impl Transaction {
    /// Creates a new transaction
    pub async fn new(
        reference: String,
        status: Status,
        typ: Type,
        spends: Vec<PaymentFrom>,
        creates: Vec<(AccountId, Amount)>,
    ) -> Result<Self, Error> {
        let creates = creates
            .into_iter()
            .map(|(to, amount)| PaymentTo { to, amount })
            .collect();

        let (transaction, revision) = BaseTx::new(spends, creates, reference, typ, status)?;
        let revision_id = revision.rev_id()?;

        Ok(Self {
            id: transaction.id()?,
            revisions: vec![revision_id.clone()],
            revision_id,
            revision,
            transaction,
        })
    }

    /// Creates a new external deposit transaction
    ///
    /// All transactions must be balanced, same amounts that are spent should be
    /// created. There are two exceptions, external deposits and withdrawals.
    /// The idea is to mimic external operations, where new assets enter the system.
    pub fn new_external_deposit(
        reference: String,
        status: Status,
        creates: Vec<(AccountId, Amount)>,
    ) -> Result<Self, Error> {
        let creates = creates
            .into_iter()
            .map(|(to, amount)| PaymentTo { to, amount })
            .collect();
        let (transaction, revision) =
            BaseTx::new(Vec::new(), creates, reference, Type::Deposit, status)?;
        let revision_id = revision.rev_id()?;

        Ok(Self {
            id: transaction.id()?,
            revisions: vec![revision_id.clone()],
            revision_id,
            revision,
            transaction,
        })
    }

    /// Creates a new external withdrawal transaction
    ///
    /// Burns assets to reflect external withdrawals
    pub fn new_external_withdrawal(
        reference: String,
        status: Status,
        spends: Vec<PaymentFrom>,
    ) -> Result<Self, Error> {
        let (transaction, revision) =
            BaseTx::new(spends, Vec::new(), reference, Type::Withdrawal, status)?;
        let revision_id = revision.rev_id()?;

        Ok(Self {
            id: transaction.id()?,
            revisions: vec![revision_id.clone()],
            revision_id,
            transaction,
            revision,
        })
    }

    /// Updates the transaction tags
    pub async fn set_tags<S>(
        self,
        config: &Config<S>,
        new_tags: Vec<String>,
        reason: String,
    ) -> Result<Self, Error>
    where
        S: Storage + Sync + Send,
    {
        let new_revision = Revision {
            transaction_id: self.revision.transaction_id,
            changelog: reason,
            previous: Some(self.revision_id),
            tags: new_tags,
            status: self.revision.status,
            created_at: Utc::now(),
        };
        let revision_id = new_revision.rev_id()?;
        let mut revisions = self.revisions;
        revisions.push(revision_id.clone());

        let mut new_transaction = Transaction {
            id: self.id,
            revisions,
            revision_id,
            transaction: self.transaction,
            revision: new_revision,
        };
        new_transaction.persist(config).await?;
        Ok(new_transaction)
    }

    /// Prepares a new revision to change the transaction status
    ///
    /// If the status transaction is not allowed, it will return an error.
    ///
    /// The new transaction with revision is returned, which is already persisted. The previous
    /// struct is consumed and the latest revision is preserved for historical purposes but it is no
    /// longer the latest revision
    #[inline]
    pub async fn change_status<S>(
        self,
        config: &Config<S>,
        new_status: Status,
        reason: String,
    ) -> Result<Self, Error>
    where
        S: Storage + Sync + Send,
    {
        config
            .status
            .is_valid_transition(&self.revision.status, &new_status)?;
        let new_revision = Revision {
            transaction_id: self.revision.transaction_id,
            changelog: reason,
            previous: Some(self.revision_id),
            tags: self.revision.tags,
            status: new_status,
            created_at: Utc::now(),
        };
        let revision_id = new_revision.rev_id()?;

        let mut new_transaction = Transaction {
            id: self.id,
            revisions: self.revisions,
            revision_id: revision_id.clone(),
            transaction: self.transaction,
            revision: new_revision,
        };
        new_transaction.revisions.push(revision_id);
        new_transaction.persist(config).await?;
        Ok(new_transaction)
    }

    fn validate(&self) -> Result<(), Error> {
        let rev_id = self.revision.rev_id()?;
        let tx_id = self.transaction.id()?;
        if self.revision_id != rev_id {
            return Err(Error::InvalidRevId(self.revision_id.clone(), rev_id));
        }
        if tx_id != self.revision.transaction_id || tx_id != self.id {
            return Err(Error::InvalidTxId(
                tx_id,
                self.revision.transaction_id.clone(),
            ));
        }
        self.transaction.validate()?;
        Ok(())
    }

    #[inline]
    async fn store_transaction<'a, S>(&mut self, batch: &mut S::Batch<'a>) -> Result<(), Error>
    where
        S: Storage + Sync + Send,
    {
        self.validate()?;
        let spends = self.spends.iter().map(|x| x.id.clone()).collect::<Vec<_>>();
        batch
            .spend_payments(
                &self.revision.transaction_id,
                spends,
                ReceivedPaymentStatus::Locked,
            )
            .await?;
        batch
            .create_payments(
                &self.revision.transaction_id,
                &self.creates,
                ReceivedPaymentStatus::Locked,
            )
            .await?;

        for account in self.accounts() {
            batch
                .relate_account_to_transaction(&self.revision.transaction_id, &account, self.typ)
                .await?;
        }
        batch
            .store_base_transaction(&self.revision.transaction_id, &self.transaction)
            .await?;
        Ok(())
    }

    /// Persists the changes done to this transaction object.
    /// This method is not idempotent, and it will fail if the transaction if the requested update
    /// is not allowed.
    pub async fn persist<'a, S>(&mut self, config: &'a Config<S>) -> Result<(), Error>
    where
        S: Storage + Sync + Send,
    {
        self.validate()?;
        let mut batch = config.storage.begin().await?;
        if self.revision.previous.is_none() {
            self.store_transaction::<S>(&mut batch).await?;
        }

        let (created_updated, spent_updated) =
            match config.status.internal_type(&self.revision.status) {
                InternalStatus::Reverted => batch
                    .update_transaction_payments(
                        &self.id,
                        ReceivedPaymentStatus::Failed,
                        ReceivedPaymentStatus::Spendable,
                    )
                    .await
                    .expect("foo0"),
                InternalStatus::Spendable => batch
                    .update_transaction_payments(
                        &self.id,
                        ReceivedPaymentStatus::Spendable,
                        ReceivedPaymentStatus::Spent,
                    )
                    .await
                    .expect("foo1"),
                _ => (self.creates.len(), self.spends.len()),
            };

        if self.creates.len() != created_updated || self.spends.len() != spent_updated {
            return Err(Error::NoUpdate);
        }

        if config.status.is_spendable(&self.revision.status) {
            batch
                .update_transaction_payments(
                    &self.id,
                    ReceivedPaymentStatus::Spendable,
                    ReceivedPaymentStatus::Spent,
                )
                .await
                .expect("foo2");
        }

        batch
            .store_revision(&self.revision_id, &self.revision)
            .await?;

        batch
            .update_transaction_revision(
                &self.id,
                &self.revision_id,
                self.revision.previous.as_ref(),
            )
            .await
            .expect("foo3");

        batch.commit().await?;

        Ok(())
    }
}
